#!/usr/bin/env python3

import boto3
import pprint
import datetime
from datetime import timedelta, timezone, datetime
from time import strftime
import dateutil.parser
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers
import logging

helpers.set_logger()
aws_client = AWS_Client()

class CleanupVolumes():

    def __init__(self, region, role_arns, not_dry_run):
          self.region = region
          self.role_arns = role_arns
          self.not_dry_run = not_dry_run

    def list_volumes(self, role_arn, filters={}):
        ec2_client = aws_client.create_client('ec2', self.region, role_arn)

        return ec2_client.describe_volumes(
            DryRun=False,
            Filters=[
                filters
            ],
            )['Volumes']

    def list_available_volumes(self, role_arn):
        return self.list_volumes(role_arn, filters={'Name': 'status', 'Values': ['available']})

    def delete_volumes(self, volumes, role_arn):
        ec2_client = aws_client.create_client('ec2', self.region, role_arn)
        volumes = self.list_available_volumes(role_arn)

        for volume in volumes:
            ec2_client.delete_volume(VolumeId=volume['VolumeId'])

    def cleanup(self):
        for role in self.role_arns:
            logging.error("Processing account: {}".format(role))
            volumes_to_delete = self.list_available_volumes(role)

            if self.not_dry_run:
                logging.info("Deleting the following volumes and their snapshots: {}".format(volumes_to_delete))
                self.delete_volumes(volumes_to_delete, role)
            else:
                logging.info("These are the volumes I would have deleted if you gave me --not-dry-run:\n")
                for volume in volumes_to_delete:
                    logging.info("Volume: {}\n".format(volume['VolumeId']))



import boto3
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers
import base64
import logging

helpers.set_logger()
aws_client = AWS_Client()

class ECRLogin():
    def __init__(self, region, assume_role_arn):
        self.region = region
        self.assume_role_arn = assume_role_arn

    def get_login(self, registry):
        ecr_client = aws_client.create_client('ecr', self.region, self.assume_role_arn)

        token = ecr_client.get_authorization_token(registryIds=[registry])['authorizationData'][0]['authorizationToken']
        decoded_token = base64.standard_b64decode(token).decode('utf-8')
        username = decoded_token.split(":")[0]
        password = decoded_token.split(":")[1]

        login_string = "docker login -u {username} -p {password} https://{registry}.dkr.ecr.{region}.amazonaws.com".format(
            username=username,
            password=password,
            registry=registry,
            region=self.region
        )

        print(login_string)



import click

# This is the Click() group that's imported into the CLI at the top level
@click.group()
@click.option("--region", required=True, help="Region your resources are located in. N.B. Currently, only us-east-1 supports estimates")
@click.option("--role-arn", required=True, help="Role ARN which contains necessary assume permissions")
@click.pass_context
def reporting(ctx, region, role_arn):
    ctx.obj = {'region': region, 'role_arn': role_arn}
    pass


@reporting.command(name="bill-estimates")
@click.pass_context
@click.option("--from-days-ago", "from_days_ago", default=0, help="Number of days ago.")
def bill_estimates(ctx, from_days_ago):
    region = ctx.obj.get('region')
    role_arn = ctx.obj.get('role_arn')

    from .billing_summary import billing_queries
    billing_queries.BillingQueries(region, role_arn).days_estimates(from_days_ago)
    
    



#!/usr/bin/env python3

import boto3
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers
import logging

helpers.set_logger()
aws_client = AWS_Client()

class CleanupSnapshots():

    def __init__(self, region, role_arns, search_tags, not_dry_run):
          self.region = region
          self.role_arns = role_arns.split(",")
          self.search_tags = search_tags.split(",")
          self.not_dry_run = not_dry_run

    def list_tagged_snapshots(self, role_arn, search_tags):
        rds_client = aws_client.create_client('rds', self.region, role_arn)

        snapshots = rds_client.describe_db_snapshots()['DBSnapshots']

        found_list = []
        for snapshot in snapshots:
            snapshot_arn = snapshot['DBSnapshotArn']
            snapshot_id = snapshot['DBSnapshotIdentifier']

            tags = rds_client.list_tags_for_resource(
                ResourceName=snapshot_arn
            )['TagList']

            for tag in tags:
                if [ snapshot_id for search_tag in search_tags if tag['Key'] == search_tag ] != []:
                    found_list.append(snapshot_id)

        return found_list

    def delete_snapshots(self, role_arn, snapshots_to_delete):
        rds_client = aws_client.create_client('rds', self.region, role_arn)

        for snapshot in snapshots_to_delete:
            try:
                logging.info(rds_client.delete_db_snapshot(DBSnapshotIdentifier=snapshot))
            except rds_client.exceptions.InvalidDBSnapshotStateFault as e:
                logging.error(
                    "Snapshot is not deletable, probably an automated snapshot:\n"
                    "{}".format(e)
                )

    def cleanup(self):
        for role in self.role_arns:
            logging.info("Processing account: {}".format(role))
            snapshots_to_delete = self.list_tagged_snapshots(role, self.search_tags)

            if self.not_dry_run:
                logging.info("Deleting the following snapshots: {}".format(snapshots_to_delete))
                self.delete_snapshots(role, snapshots_to_delete)
            else:
                logging.info("These are the snapshots I would have deleted if you gave me --not-dry-run: {}".format(snapshots_to_delete))



import akinaka.libs.cloudwatch
import akinaka.libs.costexplorer
import boto3
import time
import datetime
import sys
import tabulate
from akinaka.libs import helpers
import logging

helpers.set_logger()

class BillingQueries():
    def __init__(self, region, assume_role_arn):
        self.costexplorer = akinaka.libs.costexplorer.CostExplorer(region, assume_role_arn)

    def days_estimates(self, from_days_ago):
        try:
            response = self.costexplorer.get_bill_estimates(from_days_ago)
            data = response['ResultsByTime']
        except Exception as e:
            logging.error("Billing estimates is not available: {}".format(e))
            return e

        results = []
        if len(data) == 1:
            amount = float(data[0]['Total']['UnblendedCost']['Amount'])
            unit = data[0]['Total']['UnblendedCost']['Unit']
            date_today = datetime.datetime.now().strftime("%Y-%m-%d")
            results.append([date_today, "{} {:.2f}".format(unit, amount)])
            message = "\nToday's estimated bill\n"
        else:
            for d in data:
                unit = d['Total']['UnblendedCost']['Unit']
                amount = float(d['Total']['UnblendedCost']['Amount'])
                results.append([d['TimePeriod']['End'], "{} {:.2f}".format(unit, amount)])
            message = "\nEstimated bill for the past {} days\n".format(str(len(data)))

        message += tabulate.tabulate(results, headers=["Date", "Total"], tablefmt='psql')
        message += "\n"

        logging.info(message)
        return



from akinaka.libs import helpers
import logging

helpers.set_logger(level="ERROR")

class AkinakaGeneralError(Exception):
    pass

class AkinakaUpdateError(Exception):
    pass

class AkinakaLoggingError(Exception):
    def __init__(self, message=None):
        super().__init__(message)
        logging.error(message)

class AkinakaCriticalException(Exception):
    def __init__(self, message=None):
        super().__init__(message)
        logging.error(message)

        exit(1)



import click
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers
from time import gmtime, strftime
import logging

helpers.set_logger()
aws_client = AWS_Client()

@click.group()
@click.option("--applications", required=True, help="Comma separated list of containers to update")
@click.option("--server", required=True, help="URL that Kubernetes control plane can be reached (with protocol)")
@click.option("--token", required=True, help="Authentication token")
@click.option("--ca-file-path", required=True, help="Path to a file with the server's CA")
@click.option("--skip-auth", is_flag=True, help="Temporary flag for skipping auth while we don't need it for other commands")
@click.pass_context
def k8s(ctx, applications, server, token, ca_file_path, skip_auth):
    if skip_auth:
        ctx.obj = {'applications': applications}
    else:
        from kubernetes import client, config

        configuration = client.Configuration()
        configuration.host = server
        configuration.ssl_ca_cert = ca_file_path
        configuration.debug = True
        configuration.api_key={"authorization":"Bearer {}".format(token)}

        ctx.obj = {'configuration': configuration, 'applications': applications}


@k8s.command()
@click.pass_context
def monitor_deployment(ctx):
    applications = ctx.obj.get('applications')
    configuration = ctx.obj.get('configuration')

    from .monitor_deployment import monitor_deployment
    k8s_monitor = monitor_deployment.MonitorDeployment(configuration, applications)
    k8s_monitor.monitor_update()


@k8s.command()
@click.option("--new-image", help="Full image path, minus the tag")
@click.option("--new-tag", help="Tag to update the container's image to")
@click.option("--file-paths", required=True, help="Comma separated string with paths to deployment specs")
@click.option("--dry-run", is_flag=True, help="Flag: Write config to stdout instead of the originating file")
@click.pass_context
def update_deployment(ctx, new_image, new_tag, file_paths, dry_run):
    applications = ctx.obj.get('applications')

    from .update_deployment import update_deployment

    k8s_update = update_deployment.UpdateDeployment(applications, new_image, new_tag, file_paths, dry_run)
    k8s_update.write_new_spec()



import click
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers
from time import gmtime, strftime
import logging

@click.group()
@click.option("--region", required=True, help="Region your resources are located in")
@click.option("--role-arn", required=True, help="Role ARN which contains necessary assume permissions")
@click.pass_context
def update(ctx, region, role_arn):
    """
    Update subcommand. Does nothing by itself except pass the global options through to it's
    subcommands via ctx
    """

    ctx.obj = {'region': region, 'role_arn': role_arn, 'log_level': ctx.obj.get('log_level')}
    pass

def set_deploy_status(verb, region, role_arn, application, reset=None):
    """
    Set the status of this deploy to started by setting a timestamp in the application's
    deploying-status-APPLICATION_NAME SSM parameter
    """

    aws_client = AWS_Client()
    ssm_client = aws_client.create_client('ssm', region, role_arn)
    ssm_parameter_name = "deploying-status-{}".format(application)

    try:
        deploying_state = ssm_client.get_parameter(Name=ssm_parameter_name)['Parameter']['Value']
    except ssm_client.exceptions.ParameterNotFound:
        deploying_state = "false"

    if verb == "start" and deploying_state != "false" and reset != True:
        logging.error("Refusing to deploy, since it looks like we're already deploying at this timestamp: {}".format(deploying_state))
        exit(1)
    elif verb == "stop":
        new_state = "false"
    elif verb == "start":
        new_state = strftime("%Y%m%d%H%M%S", gmtime())

    ssm_client.put_parameter(
        Name=ssm_parameter_name,
        Description="Whether we're deploying right now",
        Value=new_state,
        Type="String",
        Overwrite=True
    )
    logging.info("Deployment status for {} updated".format(application))

@update.command()
@click.pass_context
@click.option("--application-name", required=True, help="The application name (target group) to reset unlock for re-running the 'asg' subcommand on")
def reset(ctx, application_name):
    """ Reset the deploy status of an application. Use with caution """

    region = ctx.obj.get('region')
    role_arn = ctx.obj.get('role_arn')
    set_deploy_status("stop", region, role_arn, application_name, True)

@update.command()
@click.pass_context
@click.option("--ami", required=True, help="Update ASG to this AMI")
@click.option("--lb", help="Loadbalancer to work out targetgroup from -- mutually exclusive with --asg and --target-group")
@click.option("--target-group", "target_group", help="Target Group to discover the ASG for updating. Mutually exclusive with --asg and --lb")
@click.option("--asg", "asg_name", help="ASG we're updating -- mutually exclusive with --lb and --target-group")
@click.option("--skip-status-check", "skip_status_check", is_flag=True, default=False, help="When passed, skips checking if we're already in the middle of a deploy")
def asg(ctx, ami, lb, asg_name, target_group, skip_status_check):
    """
    Update an ASG by scaling it down and up again with the new launch template configuration. Can be
    used in three different modes, the first two being geared towards blue/green deploys:

    --lb:           launch template will be worked out by querying the load balancer. Only applicable for load
                    balancers with a single ASG
    --target-group: launch template will be worked out by querying the target group (then the load balancer).
                    Used when the load balancer is attached to more than one ASG
    --asg:          For when you already know which ASG you want to update. Used for non-blue/green deploys,
                    such as ASGs for workers
    """

    if [lb, asg_name, target_group].count(None) < 2:
        logging.error("--lb, --asg, and --target-group are mutually exclusive. Please use only one")

    region = ctx.obj.get('region')
    role_arn = ctx.obj.get('role_arn')

    from .asg import update_asg

    asg = update_asg.ASG(region, role_arn)
    application = asg.get_application_name(asg=asg_name, loadbalancer=lb, target_group=target_group)

    if lb or target_group:
        if not skip_status_check:
            set_deploy_status("start", region, role_arn, application)

    asg.do_update(ami, asg=asg_name, loadbalancer=lb, target_group=target_group)
    exit(0)

@update.command()
@click.pass_context
@click.option("--new", "new_asg_target", help="The ASG we're switching the LB to (attaching this ASG to the LB's targetgroup)")
def targetgroup(ctx, new_asg_target):
    """
    Switch the load balancer to serve from a different ASG by attaching the new one and detaching the
    old one, from the target group
    """

    region = ctx.obj.get('region')
    role_arn = ctx.obj.get('role_arn')
    log_level = ctx.obj.get('log_level')

    from .targetgroup import update_targetgroup

    try:
        target_groups = update_targetgroup.TargetGroup(region, role_arn, new_asg_target, log_level)
        target_groups.switch_asg()
        # We've successfully deployed, so set the status of deploy to "false"
        set_deploy_status("stop", region, role_arn, target_groups.get_application_name())
        exit(0)
    except Exception as e:
        logging.error("{}".format(e))
        exit(1)

@update.command()
@click.pass_context
@click.option("--active-asg", "active_asg", help="Name of the currently active ASG")
@click.option("--skip-status-check", "skip_status_check", is_flag=True, default=False, help="When passed, skips checking if we're already in the middle of a deploy")
def scale_down_inactive(ctx, active_asg, skip_status_check):
    """
    Given an the name of the _active_ ASG, scale down the inactive one. Only useful for
    blue/green deploys
    """

    region = ctx.obj.get('region')
    role_arn = ctx.obj.get('role_arn')

    from .asg import update_asg

    asg = update_asg.ASG(region=region, role_arn=role_arn)

    asg.scale_down_inactive(active_asg)
    exit(0)



#!/usr/bin/env python3

from kubernetes import client

class MonitorDeployment():

    def __init__(self, configuration, applications):

        client.Configuration.set_default(configuration)

        self.core_api = client.CoreV1Api()
        self.apps_api = client.AppsV1Api()


    # TODO:
    # kubectl --certificate-authority ca.crt --server $K8S_URL --token $K8S_TOKEN apply -f full_spec.yml
    def monitor_update(self):
        result = self.apps_api.read_namespaced_deployment(namespace="dispute", name="dispute")
        print(result)



#!/usr/bin/env python3

def main():
    import click
    from time import gmtime, strftime
    from akinaka.libs import helpers
    from akinaka.update.commands import update as update_commands
    from akinaka.cleanup.commands import cleanup as cleanup_commands
    from akinaka.rds.commands import copy as copy_commands
    from akinaka.reporting.commands import reporting as reporting_commands
    from akinaka.container.commands import container as container_commands
    from akinaka.k8s.commands import k8s as k8s_commands
    from akinaka.dr.commands import dr as dr_commands


    @click.group()
    @click.option("--log-level", '-l', default="INFO", type=click.Choice(["INFO", "ERROR", "DEBUG"]), help="How much information to show in logging. Default is INFO")
    @click.pass_context
    def cli(ctx=None, log_level=None):
        ctx.obj = {'log_level': log_level}

    cli.add_command(update_commands)
    cli.add_command(cleanup_commands)
    cli.add_command(copy_commands)
    cli.add_command(reporting_commands)
    cli.add_command(container_commands)
    cli.add_command(k8s_commands)
    cli.add_command(dr_commands)

    cli()

if __name__ == "__main__":
    main()



"""
Sharing snapshots between AWS accounts involves:

1. Creating a key to share between those two accounts, and sharing it
2. (Re)encrypting a snapshot from the live account with the shared key
3. Creating a key on the destination account
4. Copying and re-encrypting the copy to the destination account with that key

This module has all the methods needed to do that, and uses them in the entrypoint
method; transfer_snapshot()
"""

#!/usr/bin/env python3

from datetime import datetime
from operator import itemgetter
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers, exceptions
import logging
import time

helpers.set_logger()
aws_client = AWS_Client()

class TransferSnapshot():
    def __init__(
            self,
            region,
            source_role_arn,
            destination_role_arn,
            source_kms_key,
            destination_kms_key
        ):

        self.region = region
        self.source_role_arn = source_role_arn
        self.destination_role_arn = destination_role_arn
        self.source_kms_key = source_kms_key
        self.destination_kms_key = destination_kms_key

    def transfer_snapshot(self, take_snapshot, db_names, source_account, destination_account, keep, retention):
        """
        For every DB in [db_names], call methods to perform the actions listed in this module's
        docstring. Additionally, rotate the oldest snapshot out, if there are more than [retention]
        """

        for db_name in db_names:
            source_rds_client = aws_client.create_client('rds', self.region, self.source_role_arn, valid_for=14400)

            if take_snapshot:
                source_snapshot = self.take_snapshot(source_rds_client, db_name, self.source_kms_key)
                logging.info("Will now recrypt it with the shared key")
            else:
                source_snapshot = self.get_latest_snapshot(db_name)

            recrypted_snapshot = self.recrypt_snapshot(source_rds_client, source_snapshot, self.source_kms_key, source_account)

            self.share_snapshot(recrypted_snapshot, destination_account)

            logging.info('The snapshot must now be recrypted and copied with a key available only to the destination account')
            destination_rds_client = aws_client.create_client('rds', self.region, self.destination_role_arn, valid_for=14400)
            self.recrypt_snapshot(destination_rds_client, recrypted_snapshot, self.destination_kms_key, destination_account)

            self.rotate_snapshots(retention, db_name, keep=None)

    def rotate_snapshots(self, retention, db_name, keep):
        """
        Get all the snapshots for [db_name], and delete the oldest one if there are more than
        [retention] of them. Ignore any in the list [keep].

        Beware, this does not take distinct days into account, only the number of snapshots. So if you
        take more than [retention] snapshots in one day, all previous snapshots will be deleted
        """

        keep = keep or []

        destination_rds_client = aws_client.create_client('rds', self.region, self.destination_role_arn, valid_for=14400)

        snapshots = destination_rds_client.describe_db_snapshots(DBInstanceIdentifier=db_name)['DBSnapshots']
        if len(snapshots) > retention:
            oldest_snapshot = sorted(snapshots, key=itemgetter('SnapshotCreateTime'))[-1]

            if oldest_snapshot['DBSnapshotIdentifier'] not in keep:
                logging.info("There are more than the given retention number of snapshots in the account," \
                    "so we're going to delete the oldest: {}".format(oldest_snapshot['DBSnapshotIdentifier'])
                )

                destination_rds_client.delete_db_snapshot(
                    DBSnapshotIdentifier=oldest_snapshot['DBSnapshotIdentifier']
                )
            else:
                logging.info("Oldest snapshot ({}) is older than" \
                             "the retention period allows, but it's " \
                             "the --keep list so it will not be deleted".format(oldest_snapshot['DBSnapshotIdentifier']))

    def get_latest_snapshot(self, db_name):
        """
        Return the latest snapshot for [db_name], where the ARN can also be the name of the DB

        Note: You can only use the db_name if you are in the account with the DB in it, else you
              must use the DB name
        """

        source_rds_client = aws_client.create_client('rds', self.region, self.source_role_arn, valid_for=14400)
        # TODO: Yuk
        #       https://stackoverflow.com/questions/59285540/rewrite-python-method-depending-on-condition
        try:
            snapshots = source_rds_client.describe_db_snapshots(DBInstanceIdentifier=db_name)['DBSnapshots']
            latest = sorted(snapshots, key=itemgetter('SnapshotCreateTime'))[0]
            logging.info("Using snapshot {}".format(latest['DBSnapshotIdentifier']))
        except IndexError:
            snapshots = source_rds_client.describe_db_cluster_snapshots(DBClusterIdentifier=db_name)['DBClusterSnapshots']
            if len(snapshots) == 0:
                raise exceptions.AkinakaCriticalException("No snapshots found for {}. You'll need to take one first with --take-snapshot".format(db_name))
            latest = sorted(snapshots, key=itemgetter('SnapshotCreateTime'))[0]
            logging.info("Using snapshot {}".format(latest['DBClusterSnapshotIdentifier']))

        return latest

    def make_snapshot_name(self, db_name, account):
        """ Make a name based on [db_name] and [account] """

        date = datetime.utcnow().strftime('%Y%m%d-%H%M')

        return "{}-{}-{}".format(db_name, date, account)

    def recrypt_snapshot(self, rds_client, snapshot, kms_key, destination_account, tags=None):
        """
        Recrypt a snapshot [snapshot] with the KMS key [kms_key]. Return the recrypted snapshot.
        """

        # TODO: Yuk
        #       https://stackoverflow.com/questions/59285540/rewrite-python-method-depending-on-condition
        try:
            new_snapshot_id = self.make_snapshot_name(snapshot['DBInstanceIdentifier'], destination_account)

            try:
                recrypted_snapshot = rds_client.copy_db_snapshot(
                    SourceDBSnapshotIdentifier=snapshot['DBSnapshotArn'],
                    TargetDBSnapshotIdentifier=new_snapshot_id,
                    KmsKeyId=kms_key['KeyMetadata']['Arn'],
                    Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ] # FIXME: Add custom tags
                )

                self.wait_for_snapshot(recrypted_snapshot['DBSnapshot'], rds_client)

                logging.info("Recrypted snapshot {} with key {}".format(
                        recrypted_snapshot['DBSnapshot']['DBSnapshotIdentifier'],
                        kms_key['KeyMetadata']['Arn']
                    ))

                return recrypted_snapshot['DBSnapshot']
            except rds_client.exceptions.DBSnapshotAlreadyExistsFault:
                snapshots = rds_client.describe_db_snapshots(DBSnapshotIdentifier=new_snapshot_id)

                logging.info("Found existing snapshot {}".format(snapshots['DBSnapshots'][0]['DBSnapshotIdentifier']))
                return snapshots['DBSnapshots'][0]

        except KeyError:
            new_snapshot_id = self.make_snapshot_name(snapshot['DBClusterIdentifier'], destination_account)

            try:
                recrypted_snapshot = rds_client.copy_db_cluster_snapshot(
                    SourceDBClusterSnapshotIdentifier=snapshot['DBClusterSnapshotArn'],
                    TargetDBClusterSnapshotIdentifier=new_snapshot_id,
                    KmsKeyId=kms_key['KeyMetadata']['Arn'],
                    Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ] # FIXME: Add custom tags
                )

                self.wait_for_snapshot(recrypted_snapshot['DBClusterSnapshot'], rds_client)

                logging.info("Recrypted snapshot {} with key {}".format(
                        recrypted_snapshot['DBClusterSnapshot']['DBClusterSnapshotIdentifier'],
                        kms_key['KeyMetadata']['Arn']
                    ))

                return recrypted_snapshot['DBClusterSnapshot']
            except rds_client.exceptions.DBClusterSnapshotAlreadyExistsFault:
                snapshots = rds_client.describe_db_cluster_snapshots(DBClusterSnapshotIdentifier=new_snapshot_id)

                logging.info("Found existing snapshot {}".format(snapshots['DBClusterSnapshots'][0]['DBClusterSnapshotIdentifier']))
                return snapshots['DBClusterSnapshots'][0]


    def share_snapshot(self, snapshot, destination_account):
        """
        Share [snapshot] with [destination_account]
        """

        # TODO: Yuk
        #       https://stackoverflow.com/questions/59285540/rewrite-python-method-depending-on-condition
        try:
            source_rds_client = aws_client.create_client('rds', self.region, self.source_role_arn, valid_for=14400)
            source_rds_client.modify_db_snapshot_attribute(
                DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'],
                AttributeName='restore',
                ValuesToAdd=[destination_account]
            )

            logging.info("Recrypted snapshot {} has been shared with account {}".format(snapshot['DBSnapshotIdentifier'], destination_account))
        except KeyError:
            source_rds_client = aws_client.create_client('rds', self.region, self.source_role_arn, valid_for=14400)
            source_rds_client.modify_db_cluster_snapshot_attribute(
                DBClusterSnapshotIdentifier=snapshot['DBClusterSnapshotIdentifier'],
                AttributeName='restore',
                ValuesToAdd=[destination_account]
            )

            logging.info("Recrypted snapshot {} has been shared with account {}".format(snapshot['DBClusterSnapshotIdentifier'], destination_account))


    def wait_for_snapshot(self, snapshot, rds_client):
        """
        Check if [snapshot] is ready by querying it every 10 seconds
        """

        # TODO: Yuk
        #       https://stackoverflow.com/questions/59285540/rewrite-python-method-depending-on-condition
        try:
            while True:
                snapshotcheck = rds_client.describe_db_snapshots(
                    DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier']
                )['DBSnapshots'][0]
                if snapshotcheck['Status'] == 'available':
                    logging.info("Snapshot {} has been created".format(snapshot['DBSnapshotIdentifier']))
                    break
                else:
                    logging.info("Snapshot {} is in progress; {}% complete".format(snapshot['DBSnapshotIdentifier'], snapshotcheck['PercentProgress']))
                    time.sleep(10)
        except KeyError:
            while True:
                snapshotcheck = rds_client.describe_db_cluster_snapshots(
                    DBClusterSnapshotIdentifier=snapshot['DBClusterSnapshotIdentifier']
                )['DBClusterSnapshots'][0]
                if snapshotcheck['Status'] == 'available':
                    logging.info("Snapshot {} has been created".format(snapshot['DBClusterSnapshotIdentifier']))
                    break
                else:
                    logging.info("Snapshot {} is in progress; {}% complete".format(snapshot['DBClusterSnapshotIdentifier'], snapshotcheck['PercentProgress']))
                    time.sleep(10)

    def take_snapshot(self, rds_client, db_name, kms_key):
        """
        Take a new snapshot of [db_name] using [kms_key]

        TODO: It's not possible to take a snapshot with a CMK, really?!
        """

        snapshot_name = self.make_snapshot_name(db_name, kms_key['KeyMetadata']['AWSAccountId'])

        try:
            snapshot = rds_client.create_db_snapshot(
                DBInstanceIdentifier=db_name,
                DBSnapshotIdentifier=snapshot_name,
                Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ]
            )

            self.wait_for_snapshot(snapshot['DBSnapshot'], rds_client)

            logging.info("Snapshot {} created".format(snapshot['DBSnapshot']['DBSnapshotIdentifier']))

            return snapshot['DBSnapshot']
        except rds_client.exceptions.DBInstanceNotFoundFault:
            snapshot = rds_client.create_db_cluster_snapshot(
                DBClusterIdentifier=db_name,
                DBClusterSnapshotIdentifier=snapshot_name,
                Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ]
            )

            self.wait_for_snapshot(snapshot['DBClusterSnapshot'], rds_client)

            logging.info("Snapshot {} created".format(snapshot['DBClusterSnapshot']['DBClusterSnapshotIdentifier']))

            return snapshot['DBClusterSnapshot']



"""
main() will:

1. Create a backup bucket in the backup account
2. Set lifecycle policies such that only [self.retention] number of versions are kept in the backup bucket
3. Set policies for the source bucket so that the backup account can get objects from it
4. Set policies for the backup bucket so that the backup account can use s3:PutBucketEncryption
5. Set an encryption policy to use [self.destination_kms_key]
6. Sync the source bucket to the backup bucket
"""

#!/usr/bin/env python3

from datetime import datetime
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers, exceptions
import logging

helpers.set_logger()
aws_client = AWS_Client()

class TransferS3():
    def __init__(
        self,
        region,
        source_role_arn,
        destination_role_arn,
        source_kms_key,
        destination_kms_key,
        retention):

        self.region = region
        self.source_role_arn = source_role_arn
        self.destination_role_arn = destination_role_arn
        self.source_kms_key = source_kms_key
        self.destination_kms_key = destination_kms_key
        self.retention = retention

    def main(self, old_bucket_names):
        """
        Go through all the actions in this module's docstring
        """

        for old_bucket_name in old_bucket_names:
            self.set_bucket_encryption(old_bucket_name, self.source_kms_key, self.source_role_arn)
            self.sync_bucket(old_bucket_name, old_bucket_name, self.source_kms_key, self.source_role_arn, self.source_role_arn)
            destination_account = self.account_id_from_role_arn(self.destination_role_arn)
            source_account = self.account_id_from_role_arn(self.source_role_arn)
            new_bucket_name = "{}-{}".format(old_bucket_name, destination_account)

            logging.info("Will create a backup bucket in the backup account if necessary")

            new_bucket_name = self.create_bucket(new_bucket_name, self.destination_role_arn)
            self.set_bucket_lifecycle(new_bucket_name, self.retention)
            self.set_bucket_policy(
                bucket=old_bucket_name,
                granter_role_arn=self.source_role_arn,
                grantee_account=destination_account
            )
            self.set_bucket_policy(
                bucket=new_bucket_name,
                granter_role_arn=self.destination_role_arn,
                grantee_account=source_account
            )
            self.set_bucket_encryption(new_bucket_name, self.destination_kms_key, self.destination_role_arn)
            self.sync_bucket(old_bucket_name, new_bucket_name, self.destination_kms_key, self.source_role_arn, self.destination_role_arn)

    def account_id_from_role_arn(self, role_arn):
        """
        Return the account ID that [role_arn] is attached to
        """

        sts_client = aws_client.create_client('sts', self.region, role_arn)
        return sts_client.get_caller_identity()['Account']

    def create_bucket(self, new_bucket_name, role_arn):
        """
        Create a bucket named [name]-backup to stored the backup objects in. Returns
        the name of the new bucket
        """

        destination_s3_client = aws_client.create_client('s3', self.region, role_arn)

        try:
            destination_s3_client.create_bucket(
                ACL='private',
                Bucket=new_bucket_name,
                CreateBucketConfiguration={
                    'LocationConstraint': self.region
                }
            )

            logging.info("Created the versioned bucket {}".format(new_bucket_name))
        except destination_s3_client.exceptions.BucketAlreadyExists:
            new_bucket_name = new_bucket_name + "-x"

            logging.info("Bucket name was taken (probably because you're trying to restore " \
                "the same account multiple times?), so the new bucket name " \
                "is going to become: {}".format(new_bucket_name))

            self.create_bucket(new_bucket_name, role_arn)
        except destination_s3_client.exceptions.BucketAlreadyOwnedByYou:
            logging.info("No need to create {}, as it already exists and we own it".format(new_bucket_name))

        destination_s3_client.put_bucket_versioning(
            Bucket=new_bucket_name,
            VersioningConfiguration= {
                'Status': 'Enabled'
            }
        )
        logging.info("Successfully applied versioning to the bucket")

        return new_bucket_name

    def set_bucket_lifecycle(self, name, retention):
        """
        Set the lifecycle policy for the versioned objects in bucket [name] to [retention] days
        """

        destination_s3_client = aws_client.create_client('s3', self.region, self.destination_role_arn)

        destination_s3_client.put_bucket_lifecycle_configuration(
            Bucket=name,
            LifecycleConfiguration={
                'Rules': [
                    {
                        'ID': 'Akinaka',
                        'Prefix': '',
                        'Status': 'Enabled',
                        'NoncurrentVersionExpiration': {
                            'NoncurrentDays': retention
                        }
                    },
                ]
            }
        )

        logging.info("Set a lifecycle policy to keep only {} " \
            "versions of objects, for the destination bucket".format(retention))

    def set_bucket_policy(self, bucket, granter_role_arn, grantee_account):
        """
        Set a policy on [bucket] such that the account of [granter_role_arn] can perform
        get, list, and put operations.

        Uses [granter_role_arn] to make the call, since
        that is the only account which already has access to make this kind of change on
        a bucket policy
        """

        source_s3_client = aws_client.create_client('s3', self.region, granter_role_arn)

        source_s3_client.put_bucket_policy(
            Bucket=bucket,
            ConfirmRemoveSelfBucketAccess=True,
            Policy="""{
                "Id": "Policy1576178812268",
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "Stmt1576178805544",
                        "Action": "*",
                        "Effect": "Allow",
                        "Resource": "arn:aws:s3:::%s/*",
                        "Principal": {
                            "AWS": [
                            "arn:aws:iam::%s:root"
                            ]
                        }
                    }
                ]
            }""" % (bucket, grantee_account)
        )

        logging.info('Successfully set a bucket policy so that account {} '\
            'can perform operations on bucket {}'.format(grantee_account, bucket))

    def set_bucket_encryption(self, bucket, kms_key, role_arn):
        """
        Set the encryption options on [bucket] to be enabled and use [kms_key]. Uses [role_arn]
        to create a client to perform the operation
        """

        s3_client = aws_client.create_client('s3', self.region, role_arn)

        s3_client.put_bucket_encryption(
            Bucket=bucket,
            ServerSideEncryptionConfiguration={
                'Rules': [
                    {
                        'ApplyServerSideEncryptionByDefault': {
                            'SSEAlgorithm': 'aws:kms',
                            'KMSMasterKeyID': kms_key['KeyMetadata']['KeyId']
                        }
                    },
                ]
            }
        )

        logging.info("Successfully set encryption on the bucket")

    def sync_bucket(self, source_bucket, destination_bucket, kms_key, source_role_arn, destination_role_arn):
        """
        Sync objects from [source_bucket] to [destination_bucket], ensuring all objects
        are encrypted with [kms_key].

        The passing of [source_role_arn] and [destination_role_arn] is so that we can (ab)use
        this method as a recryptor for when we need to restore from a backup account
        """

        source_s3_client = aws_client.create_client('s3', self.region, source_role_arn)
        destination_s3_client = aws_client.create_client('s3', self.region, destination_role_arn)

        try:
            source_objects = source_s3_client.list_objects(Bucket=source_bucket)['Contents']
        except KeyError:
            logging.error("Failed to get a listing for objects for the bucket, " \
                "probably because there were no objects to sync")
            return

        for obj in source_objects:
            copy_source = {
                'Bucket': source_bucket,
                'Key': obj['Key']
            }

            destination_s3_client.copy_object(
                ACL='private',
                Bucket=destination_bucket,
                CopySource=copy_source,
                Key=obj['Key'],
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=kms_key['KeyMetadata']['KeyId'],
            )

            logging.info("Synced object {}".format(obj['Key']))



#!/usr/bin/env python3

from akinaka.client.aws_client import AWS_Client
from akinaka.libs import exceptions
from akinaka.libs import helpers
import logging

aws_client = AWS_Client()

class TargetGroup():
    def __init__(self, region, role_arn, new_asg, log_level):
        self.region = region
        self.role_arn = role_arn
        self.new_asg = new_asg
        self.log_level = log_level
        logging.getLogger().setLevel(log_level)

    def get_application_name(self):
        asg_split = self.new_asg.split('-')[0:-1]

        return '-'.join(asg_split)

    def filter_out_unwanted_asgs(self, asgs, wanted_asg):
        """
        Takes a list of ASG objects (dicts), the ones you want to keep, and filters out everything else
        """

        return_list = []

        # Yes a regexp is smarter, but don't be smart, be safe ;)
        asg_split = wanted_asg.split('-')[0:-1]
        wanted_asg_prefix = '-'.join(asg_split)

        for asg in asgs:
            if wanted_asg_prefix in asg['AutoScalingGroupName']:
                return_list.append(asg)

        logging.debug("filter_out_unwanted_asgs(): Wanted ASG prefix: {}, Return list: {}".format(wanted_asg_prefix, return_list))

        return return_list

    def group_asgs_by_status(self, asgs, wanted_asg):
        """
        Returns a dict of 'active_asg' and 'inactive_asg' dicts with only the ASGs
        matching 'wanted_asgs'
        """

        inactive_asg = []
        active_asg = []

        for asg in asgs['AutoScalingGroups']:
            # If the ASG's TargetGroupARNs attribute is empty, it's the inactive one
            if len(asg['TargetGroupARNs']) < 1 and asg['AutoScalingGroupName'] == self.new_asg:
                inactive_asg.append(asg)

            # If the ASG's TargetGroupARNs attribute is not empty, it's the active one
            elif len(asg['TargetGroupARNs']) > 0:
                active_asg.append(asg)

        active_asg = self.filter_out_unwanted_asgs(active_asg, wanted_asg)
        inactive_asg = self.filter_out_unwanted_asgs(inactive_asg, wanted_asg)

        logging.debug("group_asgs_by_status(): active_asg: {}, inactive_asg: {} ".format(active_asg, inactive_asg))

        return {'active_asg': active_asg, 'inactive_asg': inactive_asg}

    def switch_asg(self):
        target_group_arns = []
        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
        asgs = asg_client.describe_auto_scaling_groups()

        asgs_by_status = self.group_asgs_by_status(asgs, self.new_asg)
        active_asg = asgs_by_status['active_asg']
        inactive_asg = asgs_by_status['inactive_asg']

        logging.debug("switch_asg(): asgs_by_status: {}, active_asg: {}, inactive_asg: {}".format(asgs_by_status, active_asg, inactive_asg))

        # Exit program if there is no existing active auto scaling groups to perform a switch over
        if len(inactive_asg) < 1 or len(active_asg) < 1:
            raise exceptions.AkinakaCriticalException('No auto scaling groups are available for this switch over.')

        # Get the target group ARNs from the active auto scaling group
        target_group_arns = active_asg[0]['TargetGroupARNs']

        # Add the ASG to the target group ARNs
        try:
            for asg in inactive_asg:
                asg_client.attach_load_balancer_target_groups(
                    AutoScalingGroupName=asg['AutoScalingGroupName'],
                    TargetGroupARNs=target_group_arns
                )
        except Exception as e:
            logging.error("Couldn't attach the new ASG {} to the target group {}".format(inactive_asg, target_group_arns))
            logging.error(e)
            # FIXME: Raise an exception.AkinakaCriticalException above instead of catching this
            exit(1)

        # Remove the asg from the target group
        try:
            for asg in active_asg:
                asg_client.detach_load_balancer_target_groups(
                    AutoScalingGroupName=asg['AutoScalingGroupName'],
                    TargetGroupARNs=target_group_arns
                )
        except Exception as e:
            logging.error("Couldn't attach the new ASG {} to the target group {}".format(inactive_asg, target_group_arns))
            logging.error(e)
            # FIXME: Raise an exception.AkinakaCriticalException above instead of catching this
            exit(1)

        return



#!/usr/bin/env python

from akinaka.client.aws_client import AWS_Client
import boto3
import time
import datetime
import sys
from akinaka.libs import helpers
import logging

helpers.set_logger()
aws_client = AWS_Client()

class CopyRDS():
    def __init__(self, region, source_role_arn, target_role_arn, snapshot_style, source_instance_name, overwrite_target, target_security_group, target_db_subnet, target_instance_name):
        self.region = region
        self.source_role_arn = source_role_arn
        self.target_role_arn = target_role_arn
        self.snapshot_style = snapshot_style
        self.source_instance_name = source_instance_name
        self.overwrite_target = overwrite_target
        self.target_security_group = target_security_group
        self.target_db_subnet = target_db_subnet
        self.target_instance_name = target_instance_name

    def copy_instance(self):
        logging.info("Starting RDS copy...")
        rds_source_client   = aws_client.create_client('rds', self.region, self.source_role_arn, 10800)
        rds_target_client   = aws_client.create_client('rds', self.region, self.target_role_arn, 10800)
        kms_client          = aws_client.create_client('kms', self.region, self.source_role_arn, 10800)
        source_account      = aws_client.create_client('sts', self.region, self.source_role_arn, 10800).get_caller_identity()['Account']
        target_account      = aws_client.create_client('sts', self.region, self.target_role_arn, 10800).get_caller_identity()['Account']
        target_account_arn  = aws_client.create_client('sts', self.region, self.source_role_arn, 10800).get_caller_identity()['Arn'].split('/boto')[0].replace(':sts::', ':iam::', 1).replace('assumed-role', 'role', 1)

        kms_key = self.get_kms_key(kms_client, source_account, target_account, target_account_arn)

        # Automated Amazon RDS snapshots cannot be shared with other AWS accounts.
        # To share an automated snapshot, copy the snapshot to make a manual version,
        # and then share the copy.
        # Additionally the copy needs to be re-encrypted with the Customer Managed KMS key
        if self.snapshot_style == 'running_instance':
            snapshot = self.make_snapshot_from_running_instance(rds_source_client, self.source_instance_name)
            self.wait_for_snapshot_to_be_ready(rds_source_client, snapshot)
        elif self.snapshot_style == 'latest_snapshot':
            # get latest snapshot from an AWS account with a given tag
            snapshot = self.get_latest_automatic_rds_snapshots(rds_source_client, self.source_instance_name)
        else:
            raise ValueError('snapshot_style has to be running_instance or latest_snapshot, but value {} found'.format(self.snapshot_style))

        recrypted_copy = self.recrypt_snapshot_with_new_key(rds_source_client, snapshot, kms_key)
        self.wait_for_snapshot_to_be_ready(rds_source_client, recrypted_copy)
        self.share_snapshot_with_external_account(rds_source_client, recrypted_copy, target_account)

        # an encrypted shared snapshot owned by another account cannot be restored straight up
        # so make a local copy in the target environment first
        target_copy = self.copy_shared_snapshot_to_local(rds_target_client, recrypted_copy, kms_key)

        self.wait_for_snapshot_to_be_ready(rds_target_client, target_copy)
        self.rename_or_delete_target_instance(rds_target_client, self.target_instance_name, self.overwrite_target)

        target_instance = self.create_rds_instance_from_snapshot(rds_client=rds_target_client,
                                                            snapshot=target_copy,
                                                            instancename=self.target_instance_name,
                                                            dbsubnet_group=self.target_db_subnet)

        self.wait_for_instance_to_be_ready(rds_target_client, target_instance)

        self.modify_rds_instance_security_groups(rds_client=rds_target_client, instancename=self.target_instance_name, securitygroup=self.target_security_group)

        logging.info("Finished, check instance {}!".format(self.target_instance_name))

    def get_kms_key(self, kms_client, source_account, target_account, target_account_arn):

        key_alias = 'alias/RDSBackupRestoreSharedKeyWith{}'.format(target_account)
        logging.info("Searching for Customer Managed KMS Key with alias {} that is already shared with account {}...".format(key_alias, target_account))

        # try to retrieve the KMS key with the specified alias to see if it exists
        try:
            key = kms_client.describe_key(KeyId=key_alias)
            logging.info("Found key: {}".format(key['KeyMetadata']['Arn']))
            return key
        except kms_client.exceptions.NotFoundException:
            # if it doesn't exist, create it
            logging.error("No valid key found.")
            key = self.create_shared_kms_key(kms_client, source_account, target_account, target_account_arn, key_alias)
            return key

    def create_shared_kms_key(self, kms_client, source_account, target_account, target_account_arn, key_alias):

        logging.info("Creating Customer Managed KMS Key that is shared...")

        # create a Customer Managed KMS key, needed to be able to share the encrypted snapshot
        kms_key = kms_client.create_key(
            Description="Shared encryption key with AWS account {}".format(target_account_arn),
            Policy= """{
                            "Version": "2012-10-17",
                            "Id": "key-default-1",
                            "Statement": [
                                {
                                    "Sid": "Enable IAM User Permissions",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": "arn:aws:iam::%s:root"
                                    },
                                    "Action": "kms:*",
                                    "Resource": "*"
                                },
                                {
                                    "Sid": "Allow use of the key by %s",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": "%s"
                                    },
                                    "Action": "kms:*",
                                    "Resource": "*"
                                }
                            ]
                        }""" % (source_account, target_account, target_account_arn)
        )

        # add an alias to the key so we can later more easily determine if the key
        # already exists without having to know it's keyid
        kms_client.create_alias(
            AliasName=key_alias,
            TargetKeyId=kms_key['KeyMetadata']['Arn']
        )

        logging.info("Created KMS Key {}, shared with account {}".format(kms_key['KeyMetadata']['Arn'], target_account_arn))
        return kms_key

    def copy_shared_snapshot_to_local(self, rds_client, shared_snapshot, kms_key):
        # unfortunately it's not possible to restore an RDS instance directly from a
        # snapshot that is shared by another account. This makes a copy local to the
        # account where we want to restore the RDS instance
        target_db_snapshot_id = "{}-copy".format(shared_snapshot['DBSnapshotIdentifier'])

        logging.info("Copying shared snaphot {} to local snapshot {}...".format(shared_snapshot['DBSnapshotArn'], target_db_snapshot_id))

        try:
            copy = rds_client.copy_db_snapshot(
                SourceDBSnapshotIdentifier=shared_snapshot['DBSnapshotArn'],
                TargetDBSnapshotIdentifier=target_db_snapshot_id,
                KmsKeyId=kms_key['KeyMetadata']['Arn'],
                Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ]
            )
            logging.info("Copy created.")
            return copy['DBSnapshot']
        except rds_client.exceptions.DBSnapshotAlreadyExistsFault:
            # if the snapshot we tried to make already exists, retrieve it
            logging.info("Snapshot already exists, retrieving {}...".format(target_db_snapshot_id))

            snapshots = rds_client.describe_db_snapshots(
                DBSnapshotIdentifier=target_db_snapshot_id,
            )
            logging.info("Retrieved.")
            return snapshots['DBSnapshots'][0]

    def share_snapshot_with_external_account(self, rds_client, snapshot, target_account):
        # in order to restore a snapshot from another account it needs to be shared
        # with that account first
        logging.info("Modifying snaphot {} to be shared with account {}...".format(snapshot['DBSnapshotArn'], target_account))
        rds_client.modify_db_snapshot_attribute(
            DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'],
            AttributeName='restore',
            ValuesToAdd=[target_account]
        )
        logging.info("Modified.")

    def rename_or_delete_target_instance(self, rds_client, instancename, overwrite_target ):
        logging.info("Checking for an existing RDS instance by the name {} and renaming or deleting if it's found".format(instancename))
        # check if we already have an instance by this name
        try:
            instance = rds_client.describe_db_instances(DBInstanceIdentifier=instancename)['DBInstances'][0]
            logging.info("Instance found")
        except rds_client.exceptions.DBInstanceNotFoundFault:
            instance = None
            logging.info("Instance not found")

        if instance is not None:
            if overwrite_target:
                logging.info("Instance found and overwrite if found True, deleting instance")
                rds_client.delete_db_instance(
                    DBInstanceIdentifier=instancename,
                    SkipFinalSnapshot=True
                    )
                logging.info("Deleting instance. This will take a while...")
                waiter = rds_client.get_waiter('db_instance_deleted')
                waiter.wait(
                    DBInstanceIdentifier=instancename,
                    WaiterConfig={
                        'MaxAttempts': 120
                    }
                )
                logging.info("Instance is deleted!")
            else:
                logging.info("Instance found and renaming instance")
                try:
                    rds_client.modify_db_instance(
                        DBInstanceIdentifier=instancename,
                        NewDBInstanceIdentifier="{}-old".format(instancename),
                        ApplyImmediately=True
                    )
                except:
                    raise

    def wait_for_instance_to_be_ready(self, rds_client, instance):
        # simply check if the specified instance is healthy every 5 seconds until it
        # is
        while True:
            instancecheck = rds_client.describe_db_instances(DBInstanceIdentifier=instance['DBInstance']['DBInstanceIdentifier'])['DBInstances'][0]
            if instancecheck['DBInstanceStatus'] == 'available':
                logging.info("Instance {} ready and available!".format(instance['DBInstance']['DBInstanceIdentifier']))
                break
            else:
                logging.info("Instance creation in progress, sleeping 10 seconds...")
                time.sleep(10)

    def wait_for_snapshot_to_be_ready(self, rds_client, snapshot):
        # simply check if the specified snapshot is healthy every 5 seconds until it
        # is
        while True:
            snapshotcheck = rds_client.describe_db_snapshots(DBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'])['DBSnapshots'][0]
            if snapshotcheck['Status'] == 'available':
                logging.info("Snapshot {} complete and available!".format(snapshot['DBSnapshotIdentifier']))
                break
            else:
                logging.info("Snapshot {} in progress, {}% complete".format(snapshot['DBSnapshotIdentifier'], snapshotcheck['PercentProgress']))
                time.sleep(10)

    def make_snapshot_from_running_instance(self, rds_client, source_instance_name):
        logging.info("Making a new snapshot from the running RDS instance")
        try:
            today = datetime.date.today()
            snapshot = rds_client.create_db_snapshot(
                DBInstanceIdentifier=source_instance_name,
                DBSnapshotIdentifier="{}-{:%Y-%m-%d}".format(source_instance_name, today),
                Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ]
            )
            logging.info("Snapshot created.")
            return snapshot['DBSnapshot']
        except Exception as exception:
            logging.error("Failed to make snapshot from instance: {}".format(exception))
            sys.exit(1)

    def get_latest_automatic_rds_snapshots(self, rds_client, source_instance_name):
        logging.info("Getting latest (automated) snapshot from rds instance {}...".format(source_instance_name))
        # we can't query for the latest snapshot straight away, so we have to retrieve
        # a full list and go through all of them
        snapshots = rds_client.describe_db_snapshots(
            DBInstanceIdentifier=source_instance_name,
            SnapshotType='automated'
        )

        latest = 0
        for snapshot in snapshots['DBSnapshots']:
            if latest == 0:
                latest = snapshot
            if snapshot['SnapshotCreateTime'] > latest['SnapshotCreateTime']:
                latest = snapshot

        logging.info("Found snapshot {}".format(latest['DBSnapshotIdentifier']))
        return latest

    def recrypt_snapshot_with_new_key(self, rds_client, snapshot, kms_key):
        # create an identifier to use as the name of the manual snapshot copy
        if ':' in snapshot['DBSnapshotIdentifier']:
            target_db_snapshot_id = "{}-recrypted".format(snapshot['DBSnapshotIdentifier'].split(':')[1])
        else:
            target_db_snapshot_id = "{}-recrypted".format(snapshot['DBSnapshotIdentifier'])

        logging.info("Copying automatic snapshot to manual snapshot...")

        try:
            # copy the snapshot, supplying the new KMS key (which is also shared with
            # the target account)
            copy = rds_client.copy_db_snapshot(
                SourceDBSnapshotIdentifier=snapshot['DBSnapshotIdentifier'],
                TargetDBSnapshotIdentifier=target_db_snapshot_id,
                KmsKeyId=kms_key['KeyMetadata']['Arn'],
                Tags=[ { 'Key': 'akinaka-made', 'Value': 'true' }, ]
            )
            logging.info("Snapshot created.")
            return copy['DBSnapshot']
        except rds_client.exceptions.DBSnapshotAlreadyExistsFault:
            # if the snapshot we tried to make already exists, retrieve it
            logging.info("Snapshot already exists, retrieving {}".format(target_db_snapshot_id))

            snapshots = rds_client.describe_db_snapshots(
                DBSnapshotIdentifier=target_db_snapshot_id,
            )

            return snapshots['DBSnapshots'][0]

    def create_rds_instance_from_snapshot(self, rds_client, snapshot, instancename, dbsubnet_group):
        # restore an instance from the specified snapshot
        logging.info("Restoring RDS instance {} from snapshot {}".format(instancename, snapshot['DBSnapshotIdentifier']))
        try:
            if dbsubnet_group is None:
                dbsubnet_group = 'default'

            instance = rds_client.restore_db_instance_from_db_snapshot(
                DBInstanceIdentifier=instancename,
                DBSnapshotIdentifier=snapshot['DBSnapshotArn'],
                DBSubnetGroupName=dbsubnet_group,
            )
            logging.info("RDS instance restored.")
            return instance
        except rds_client.exceptions.DBInstanceAlreadyExistsFault:
            logging.error("An instance with the name {} already exists, please specify a different name or remove that instance first".format(instancename))
            sys.exit(1)

    def modify_rds_instance_security_groups(self, rds_client, instancename, securitygroup):
        logging.info("Modifying RDS instance to attach correct securitygroup")
        try:
            rds_client.modify_db_instance(
                DBInstanceIdentifier=instancename,
                VpcSecurityGroupIds=[
                    securitygroup
                ],
                ApplyImmediately=True
            )
            logging.info("RDS Instance {} modified".format(instancename))
        except Exception as e:
            logging.error("{}".format(e))
            raise



#!/usr/bin/env python3

import boto3
import pprint
import datetime
from datetime import timedelta, timezone, datetime
from time import strftime
import dateutil.parser
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers
import logging

helpers.set_logger()
aws_client = AWS_Client()

class CleanupAMIs():

    def __init__(self, region, role_arns, retention, not_dry_run, exceptional_amis=None, launch_templates=None):
          self.region = region
          self.role_arns = role_arns
          self.retention = int(retention)
          self.exceptional_amis = exceptional_amis
          self.launch_templates = launch_templates
          self.retention_end = (datetime.now(timezone.utc) + timedelta(days=-self.retention))
          self.not_dry_run = not_dry_run

    def list_amis(self, filters={}, ids=[]):
        ec2_client = aws_client.create_client('ec2', self.region, self.role_arns[0])
        sts_client = aws_client.create_client('sts', self.region, self.role_arns[0])
        account_id = sts_client.get_caller_identity().get('Account')

        return ec2_client.describe_images(
            Owners=[ account_id ],
            DryRun=False,
            Filters=[
                filters
            ],
            ImageIds=ids
            )['Images']

    def delist_out_of_retention_amis(self, all_amis):
        amis_to_delete = set()

        for ami in all_amis:
            if dateutil.parser.parse(ami['CreationDate']) < self.retention_end:
                amis_to_delete.add(ami['ImageId'])

        # Same with list comprehension
        # logging.info([ami for ami in all_amis if dateutil.parser.parse(ami['CreationDate']) < self.retention_start])
        return amis_to_delete

    def delist_in_use_amis(self, amis):
        unused_amis = set()
        all_accounts_instances = []
        instance_amis = set()

        for role in self.role_arns:
            ec2_client = aws_client.create_client('ec2', self.region, role)
            all_accounts_instances.append(ec2_client.describe_instances()['Reservations'])

        all_accounts_instances = [account for accounts in all_accounts_instances for account in accounts]
        for account_instance in all_accounts_instances:
            instance_amis.add(account_instance['Instances'][0]['ImageId'])

        for ami in amis:
            if ami in instance_amis:
                continue
            unused_amis.add(ami)

        return unused_amis

    def delist_latest_arbitrary_amis(self, amis):
        for exceptional_ami in self.exceptional_amis:

            filter = {
                'Name': 'name',
                'Values': [ exceptional_ami ]
            }

            all_arbitrary_amis = self.list_amis(filter)
            amis.discard(max(all_arbitrary_amis, key=lambda x:x['CreationDate'])['ImageId'])

        return amis

    def delist_launch_template_finds(self, amis):
        ec2_client = aws_client.create_client('ec2', self.region, self.role_arns[0])

        for launch_template in self.launch_templates:
            latest_version = ec2_client.describe_launch_templates(
                LaunchTemplateNames=[launch_template]
            )['LaunchTemplates'][0]['LatestVersionNumber']
            str(latest_version)

            # It shouldn't be possible to have no value for ImageId, but it is :D
            try:
                in_use_ami = ec2_client.describe_launch_template_versions(
                    LaunchTemplateName=launch_template,
                    Versions=[latest_version]
                )['LaunchTemplateVersions'][0]['LaunchTemplateData']['ImageId']
            except:
                in_use_ami = ""

            amis.discard(in_use_ami)

        return amis

    def get_snapshot(self, ami):
        return self.list_amis(ids=[ami])[0]['BlockDeviceMappings'][0]['Ebs']['SnapshotId']

    def delete_amis(self, amis):
        ec2_client = aws_client.create_client('ec2', self.region, self.role_arns[0])

        for ami in amis:
            snapshot = self.get_snapshot(ami)
            ec2_client.deregister_image(ImageId=ami)
            ec2_client.delete_snapshot(SnapshotId=snapshot)

    def cleanup(self):
        amis_to_delete = self.delist_out_of_retention_amis(self.list_amis())
        amis_to_delete = self.delist_in_use_amis(amis_to_delete)
        amis_to_delete = self.delist_latest_arbitrary_amis(amis_to_delete)
        amis_to_delete = self.delist_launch_template_finds(amis_to_delete)

        if self.not_dry_run:
            logging.info("Deleting the following AMIs and their snapshots: {}".format(amis_to_delete))
            self.delete_amis(amis_to_delete)
        else:
            logging.info("These are the AMIs I would have deleted if you gave me --not-dry-run: {}".format(amis_to_delete))



import boto3
from akinaka.client.aws_client import AWS_Client
import akinaka.libs.helpers as helpers
from datetime import datetime, timezone

aws_client = AWS_Client()

class CloudWatch():
    def __init__(self, region, role_arn):
        self.cloudwatch_client = aws_client.create_client('cloudwatch', region, role_arn)

    def get_metric_statistics(self, namespace, name, seconds_ago=None, granularity=None, fields=None, stat_types=None, start=None, end=None):
        granularity = granularity or 120
        seconds_ago = seconds_ago or 86400
        start = start or helpers.datetime_this_seconds_ago(seconds_ago)
        end = end or datetime.now(timezone.utc)

        try:
            return self.cloudwatch_client.get_metric_statistics(
                Namespace = namespace,
                MetricName = name,
                StartTime = start,
                EndTime = end,
                Period = granularity,
                Dimensions = fields,
                Statistics = stat_types
            )
        except Exception as e:
            return "Too many data points to return. Try adjusting either \"seconds_ago\" or \"granularity\\n{}".format(e)

    def get_bill_estimates(self, from_hours_ago=None):
        seconds_ago = None or helpers.seconds_from_hours(from_hours_ago)

        return self.get_metric_statistics(
            namespace = "AWS/Billing",
            name = "EstimatedCharges",
            seconds_ago = seconds_ago,
            granularity = 600,
            fields = [{ 'Name': 'Currency', 'Value': 'USD' }],
            stat_types = ["Maximum"],
        )['Datapoints']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Create shared KMS keys for use in encrypting and decrypting the same objects in different accounts.

Terminology used below:

share_from_account = The account to create the key in, where the resources to be encrypted are
share_to_account = The account you're calling this from
assumable_role_arn = Role in share_from_account that permits STS assume from share_to_account
"""

import boto3
from akinaka.libs import helpers
import logging
from akinaka.client.aws_client import AWS_Client

aws_client = AWS_Client()
helpers.set_logger()

class KMSShare():
    def __init__(self, region, assumable_role_arn, share_from_account, share_to_account):
        self.region = region
        self.assumable_role_arn = assumable_role_arn
        self.share_from_account = share_from_account
        self.share_to_account = share_to_account
        self.share_to_account_arn = "arn:aws:iam::{}:root".format(share_to_account)

    def get_kms_key(self, share_from_account):
        """
        Check to see if there is already a shared key. The name (alias) to check will be:
        alias/RDSBackupRestoreSharedKeyWith[(account id of)share_to_account]. If there is no shared key,
        call create_shared_key() to make one.

        In both cases, return the ID of a shared KMS key
        """

        live_kms_client = aws_client.create_client('kms', self.region, self.assumable_role_arn)

        key_alias = 'alias/SharedKeyWithAccount{}'.format(self.share_to_account)
        logging.info("Searching for Customer Managed KMS Key with alias {} that is already shared with account {}".format(key_alias, self.share_to_account))

        try:
            key = live_kms_client.describe_key(KeyId=key_alias)
            logging.info("Found key: {}".format(key['KeyMetadata']['Arn']))

            return key
        except live_kms_client.exceptions.NotFoundException:
            logging.info("No valid key found")
            key = self.create_shared_key(share_from_account, key_alias)

            return key

    def create_shared_key(self, share_from_account, key_alias):
        """
        Create new KMS key in share_from_account, with a policy that shares it
        to share_to_account (the account you're calling this from)
        """

        logging.info("Creating a shared KMS key")

        live_kms_client = aws_client.create_client('kms', self.region, self.assumable_role_arn)

        key_policy = """{
                            "Version": "2012-10-17",
                            "Id": "key-default-1",
                            "Statement": [
                                {
                                    "Sid": "Enable IAM User Permissions",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": "arn:aws:iam::%s:root"
                                    },
                                    "Action": "kms:*",
                                    "Resource": "*"
                                },
                                {
                                    "Sid": "Allow use of the key by the %s account",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": "%s"
                                    },
                                    "Action": "kms:*",
                                    "Resource": "*"
                                }
                            ]
                        }""" % (self.share_from_account, self.share_to_account, self.share_to_account_arn)

        kms_key = live_kms_client.create_key(
            Description="Shared encryption key with AWS account {}".format(self.share_to_account),
            Policy=key_policy)

        live_kms_client.create_alias(
            AliasName=key_alias,
            TargetKeyId=kms_key['KeyMetadata']['Arn']
        )

        logging.info("Created KMS Key {}, shared with account {}".format(kms_key['KeyMetadata']['Arn'], self.share_to_account))

        return kms_key



import click
from akinaka.libs import helpers
import logging
from akinaka.cleanup.rds import cleanup_snapshots
helpers.set_logger()

@click.group()
@click.option("--region", envvar='AWS_DEFAULT_REGION', help="Region your resources are located in")

@click.pass_context

# This is the Click() group that's imported into the CLI at the top level
def copy(ctx, region):
    ctx.obj = {'region': region}
    pass


@copy.command()
@click.pass_context
@click.option("--source-role-arn", required=True, help="Source role ARNs with assumable permissions")
@click.option("--target-role-arn", required=True, help="Destination role ARNs with assumable permissions")
@click.option("--snapshot-style", type=click.Choice(['running_instance', 'latest_snapshot']), required=True, help="Use latest available backup or create a new snapshot")
@click.option("--source-instance-name", required=True, help="RDS DB instance identifier")
@click.option("--target-instance-name", required=True, default=None, help="Name of the newly created RDS instance")
@click.option("--overwrite-target", is_flag=True, help="Specify this parameter to overwrite existing instance")
@click.option("--target-security-group", required=True, help="RDS Security to be attached to the target RDS instance")
@click.option("--target-db-subnet", required=True, help="RDS DB subnet to be attached to the instance")
def rds(ctx, source_role_arn, target_role_arn, snapshot_style, source_instance_name, overwrite_target, target_security_group, target_db_subnet, target_instance_name):
    from .copy import copy_rds
    region = ctx.obj.get('region')

    try:
        rds_copy = copy_rds.CopyRDS(region,
                                    source_role_arn,
                                    target_role_arn,
                                    snapshot_style,
                                    source_instance_name,
                                    overwrite_target,
                                    target_security_group,
                                    target_db_subnet,
                                    target_instance_name)
        rds_copy.copy_instance()

        logging.info("Will now delete useless snapshots")
        snapshots = cleanup_snapshots.CleanupSnapshots(
            region,
            "{},{}".format(source_role_arn,target_role_arn),
            "akinaka-made",
            True
        )
        snapshots.cleanup()

        exit(0)
    except Exception as e:
        logging.error(e)
        exit(1)



#!/usr/bin/env python3

from time import sleep
from akinaka.libs import helpers
from akinaka.libs import exceptions
from botocore.exceptions import ParamValidationError
from botocore.exceptions import ClientError
from akinaka.client.aws_client import AWS_Client
import logging

aws_client = AWS_Client()

class ASG():
    """All the methods needed to perform a blue/green deploy"""

    def __init__(self, region, role_arn):
        self.region = region
        self.role_arn = role_arn

    def get_application_name(self, asg, loadbalancer=None, target_group=None):
        """
        Returns the application name that we're deploying, worked out from the target group
        (via the load balancer if applicable)
        """

        if asg:
            return asg

        target_group_arn = self.get_target_group_arn(loadbalancer, target_group)
        active_asg = self.get_active_asg(target_group_arn)
        asg_split = active_asg.split('-')[0:-1]

        return '-'.join(asg_split)

    def asgs_by_liveness(self, asg=None, loadbalancer=None, target_group=None):
        """Return dict of '{inactive: ASG, active: ASG}'"""

        if asg is not None:
            logging.info("We've been given the ASG name as an argument")
            # NOTE: "inactive_asg" is a misnomer at this point, since when we already have the ASG
            #       name, it is also the active ASG, because this case is used for non-blue/green ASGs
            return {"inactive_asg": asg, "active_asg": asg}

        target_group_arn = self.get_target_group_arn(loadbalancer=loadbalancer, target_group=target_group)
        active_asg = self.get_active_asg(target_group_arn)
        inactive_asg = self.get_inactive_asg(active_asg)

        return {"inactive_asg": inactive_asg, "active_asg": active_asg}

    def scale_down_inactive(self, asg):
        # NOTE: We're making the heavy presumption here that the _active_ ASG is the one we've been
        #       given from the command line
        inactive_asg = self.get_inactive_asg(asg)
        self.scale(inactive_asg, 0, 0, 0)

    def set_asg_launch_template_version(self, asg, lt_id, lt_version):
        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)

        asg_client.update_auto_scaling_group(
            AutoScalingGroupName = asg,
            LaunchTemplate = {
                "LaunchTemplateId": lt_id,
                "Version": lt_version
            }
        )

    def do_update(self, ami, asg=None, loadbalancer=None, target_group=None):
        asg_liveness_info = self.asgs_by_liveness(asg=asg, loadbalancer=loadbalancer, target_group=target_group)

        inactive_asg = asg_liveness_info['inactive_asg']
        active_asg = asg_liveness_info['active_asg']
        new_ami = ami
        current_lt_info = self.get_launch_template_info(self.get_lt_name(inactive_asg))

        logging.info("New ASG was worked out as {}. Now updating it's Launch Template".format(inactive_asg))

        # Set the current ASG to the current launch template version. Shouldn't actually be necessary,
        # since it's done during the previous deploy, but including for completeness and paranoia
        self.set_asg_launch_template_version(
            asg=active_asg,
            lt_id=current_lt_info["id"],
            lt_version=current_lt_info["version"],
        )

        # Update the lt and set the soon to be new ASG to the new launch template version
        updated_lt = self.update_launch_template(new_ami, self.get_lt_name(inactive_asg))
        self.set_asg_launch_template_version(
            asg=inactive_asg,
            lt_id=updated_lt["id"],
            lt_version=updated_lt["version"]
        )

        scale_to = self.get_current_scale(active_asg)

        logging.info("Scaling ASG '{}' down to ({}, {}, {})".format(inactive_asg, 0, 0, 0))
        self.scale(inactive_asg, 0, 0, 0)
        while not self.asg_is_empty(inactive_asg):
            logging.info("Waiting for instances in ASG '{}' to terminate".format(inactive_asg))
            sleep(10)

        logging.info("Scaling ASG '{}' back up to ({}, {}, {})".format(inactive_asg, scale_to['min'], scale_to['max'], scale_to['desired']))
        self.scale(
            auto_scaling_group_id = inactive_asg,
            min_size = scale_to['min'],
            max_size = scale_to['max'],
            desired =  scale_to['desired']
        )

        while len(self.get_auto_scaling_group_instances(inactive_asg)) < 1:
            logging.info("Waiting for instances in ASG to start ...")
            sleep(10)

        logging.info("First instance has started")

        # Try to get information for an instance in the new ASG 20 times
        for i in range(20):
            try:
                first_new_instance = self.get_first_new_instance(inactive_asg)
                if i == 20:
                    raise exceptions.AkinakaCriticalException
                break
            except exceptions.AkinakaLoggingError as error:
                logging.info("Retry {}".format(i))
                logging.info("Problem in getting data for first new ASG instance")
                logging.info("get_auto_scaling_group_instances() returned: {}".format(self.get_auto_scaling_group_instances(inactive_asg)))
                logging.info("Error was: {}".format(error))

        # Show console output for first instance up until it's Lifecycle Hook has passed
        while self.get_auto_scaling_group_instances(auto_scaling_group_id=inactive_asg, instance_ids=[first_new_instance])[0]['LifecycleState'] != "InService":
            try:
                logging.info("Attempting to retrieve console output from first instance up -- this will not work for non-nitro hypervisor VMs")
                logging.info(self.get_instance_console_output(first_new_instance)['Output'])
            except KeyError:
                logging.info("No output from instance yet. Trying again in 10 seconds.")
            sleep(10)

        # Wait for remaining instances (if any) to come up too
        while len(self.asgs_healthy_instances(inactive_asg)) < scale_to['desired']:
            logging.info("Waiting for all instances to be healthy ...")

        logging.info("ASG fully healthy. Logging new ASG name to \"inactive_asg.txt\"")
        open("inactive_asg.txt", "w").write(inactive_asg)

    def get_first_new_instance(self, new_asg):
        asg_instances = self.get_auto_scaling_group_instances(new_asg)
        if len(asg_instances) < 1: raise exceptions.AkinakaLoggingError

        return asg_instances[0]['InstanceId']

    def get_instance_console_output(self, instance):
        ec2_client = aws_client.create_client('ec2', self.region, self.role_arn)
        return ec2_client.get_console_output(InstanceId=instance)

    def instance_state(self, instance):
        ec2_client = aws_client.create_client('ec2', self.region, self.role_arn)
        instance_states = ec2_client.describe_instance_status(IncludeAllInstances=True, InstanceIds=[instance])

        while 'InstanceStatuses' not in instance_states:
            instance_states = ec2_client.describe_instance_status(InstanceIds=[instance])

        return instance_states['InstanceStatuses'][0]['InstanceState']['Name']

    def get_lt_name(self, asg):
        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
        lt_info = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg])

        try:
            return lt_info['AutoScalingGroups'][0]['LaunchTemplate']['LaunchTemplateName']
        except Exception as e:
            raise exceptions.AkinakaCriticalException("{}: Likely couldn't find the ASG you're trying to update".format(e))

    def get_target_group_arn(self, loadbalancer=None, target_group=None):
        """
        Returns a string containing the ARN of the target group, either by using:

        * targetgroup from --targetgroup
        * loadbalancer from --lb

        Both are mutually exclusive, and at least one must be supplied
        """

        alb_client = aws_client.create_client('elbv2', self.region, self.role_arn)
        if target_group is not None:
            return alb_client.describe_target_groups(Names=[target_group])['TargetGroups'][0]['TargetGroupArn']

        loadbalancer_raw_info = alb_client.describe_load_balancers(Names=[loadbalancer])
        loadbalancer_arn = loadbalancer_raw_info['LoadBalancers'][0]['LoadBalancerArn']

        target_groups_raw_info = alb_client.describe_target_groups(LoadBalancerArn=loadbalancer_arn)['TargetGroups']
        target_group_arns = [targetgroup['TargetGroupArn'] for targetgroup in target_groups_raw_info]

        # If we get this far, then the LB has more than a single target group, and we can't work
        # out which one the caller wants
        if len(target_group_arns) == 1:
            return target_group_arns[0]
        elif len(target_group_arns) > 1:
            error_message = "Load balancer has {} target groups".format(len(target_group_arns))

            for target_group in target_group_arns:
                error_message += "> " + target_group
        else:
            error_message = "Load balancer has no target groups"

        raise exceptions.AkinakaCriticalException(error_message)

    def get_target_groups_instances(self, target_group_arn):
        """Returns an array of instance IDs belonging to the target group specified"""
        alb_client = aws_client.create_client('elbv2', self.region, self.role_arn)

        target_groups_instances = []

        these_target_group_instances = alb_client.describe_target_health(TargetGroupArn=target_group_arn)['TargetHealthDescriptions']
        these_target_group_instances = [
            instance for instance in these_target_group_instances if not instance['TargetHealth']['State'] == "unused"
        ]

        # NOTE: This presumes some robustness from your target groups. If they contain instances
        #       from stale ASGs, or anything else is off in them, you will get unexpected results
        for instance in these_target_group_instances:
            target_groups_instances.append(instance['Target']['Id'])

        return target_groups_instances

    def get_active_asg(self, target_groups):
        ec2_client = aws_client.create_client('ec2', self.region, self.role_arn)

        target_groups_instances = self.get_target_groups_instances(target_groups)

        instances_with_tags = {}

        raw_instances_reservations = ec2_client.describe_instances(InstanceIds=target_groups_instances)['Reservations']

        for reservation in raw_instances_reservations:
            for instance in reservation['Instances']:
                instances_with_tags[instance['InstanceId']] = instance['Tags']

        instances_with_asg = {}

        # Create dictionary with Instance -> AG
        for (key,value) in instances_with_tags.items():
            for tag in value:

                if tag['Key'] == 'aws:autoscaling:groupName':
                    instances_with_asg[key] = tag['Value']

        return next(iter(instances_with_asg.values()))

    def get_inactive_asg(self, active_asg):
        asg_parts = active_asg.split('-')
        active_colour = asg_parts[-1]

        if active_colour == "blue":
            inactive_colour = "green"
        else:
            inactive_colour = "blue"

        asg_parts[-1] = inactive_colour

        return "-".join(asg_parts)

    def get_launch_template_info(self, lt_name):
        """Returns the [id] and (latest) [version] number for lt_name"""

        ec2_client = aws_client.create_client('ec2', self.region, self.role_arn)

        response = ec2_client.describe_launch_templates(
            DryRun=False,
            Filters=[
                {
                    'Name': 'launch-template-name',
                    'Values': [ lt_name ]
                }
            ]
        )

        # FIXME: This is hackish because we are assuming that there is going to be only 1 (one) launch template
        return {
            "id": response['LaunchTemplates'][0]['LaunchTemplateId'],
            "version": str(response['LaunchTemplates'][0]['LatestVersionNumber'])
        }

    def update_launch_template(self, ami, lt_name):
        """
        Creates a new template version for [lt_name] which uses [ami], and sets the default
        default template version to be that version.

        Returns [id] and (new) [version] number
        """

        lt = self.get_launch_template_info(lt_name)
        lt_client  = aws_client.create_client('ec2', self.region, self.role_arn)

        response = lt_client.create_launch_template_version(
            DryRun = False,
            LaunchTemplateId = lt['id'],
            SourceVersion = str(lt['version']),
            LaunchTemplateData = {
                "ImageId": ami
            }
        )

        launch_template_new_version = str(response['LaunchTemplateVersion']['VersionNumber'])

        lt_client.modify_launch_template(
            LaunchTemplateId=lt['id'],
            DefaultVersion=launch_template_new_version
        )

        return {
            "id": lt['id'],
            "version": launch_template_new_version
        }

    def scale(self, auto_scaling_group_id, min_size, max_size, desired):
        """Scale an ASG to {'min_size', 'max_size', 'desired'}"""

        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)

        response = asg_client.update_auto_scaling_group(
            AutoScalingGroupName=auto_scaling_group_id,
            MinSize=min_size,
            MaxSize=max_size,
            DesiredCapacity=desired
        )

        return response

    def get_current_scale(self, asg):
        """
        Returns the current scales of the given ASG as dict {'desired', 'min', 'max'},
        expects ASG ID as argument
        """

        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
        asg = asg_client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg])['AutoScalingGroups'][0]

        return {
            "desired": asg['DesiredCapacity'],
            "min": asg['MinSize'],
            "max": asg['MaxSize']
        }

    def get_auto_scaling_group_instances(self, auto_scaling_group_id, instance_ids=None):
        asg_client = aws_client.create_client('autoscaling', self.region, self.role_arn)
        instance_ids = instance_ids if instance_ids else []
        asg_instances = asg_client.describe_auto_scaling_instances(InstanceIds=instance_ids)
        target_instances = []

        for i in asg_instances['AutoScalingInstances']:
            if i['AutoScalingGroupName'] == auto_scaling_group_id:
                target_instances.append(i)

                logging.debug("Instance {instance_id} has state = {instance_state},"
                    "Lifecycle is at {instance_lifecycle_state}".format(
                        instance_id = i['InstanceId'],
                        instance_state = i['HealthStatus'],
                        instance_lifecycle_state = i['LifecycleState']
                    )
                )

                logging.info("Instance {instance_id} is {instance_lifecycle_state}".format(
                        instance_id = i['InstanceId'],
                        instance_lifecycle_state = i['LifecycleState']
                    )
                )


        return target_instances

    def asg_is_empty(self, auto_scaling_group_id):
        asg_instances = self.get_auto_scaling_group_instances(auto_scaling_group_id)

        return all(instance.get('LifecycleState') == 'Terminated' for instance in asg_instances)

    def asgs_healthy_instances(self, auto_scaling_group_id):
        asg_instances = self.get_auto_scaling_group_instances(auto_scaling_group_id)
        healthy_instances = []

        for i in asg_instances:
            if i['LifecycleState'] == "InService":
                healthy_instances.append(i)

        return healthy_instances



import click
from akinaka.libs import helpers
import logging

helpers.set_logger()

@click.group()
@click.option("--region", required=True, help="Region your resources are located in")
@click.option("--role-arns", required=True, help="Role ARNs with assumable permissions, to do the cleanup for")
@click.option("--not-dry-run", is_flag=True, help="Will do nothing unless supplied")
@click.pass_context
def cleanup(ctx, region, role_arns, not_dry_run=False):
    ctx.obj = {'region': region, 'role_arns': role_arns, 'not_dry_run': not_dry_run}
    pass


@cleanup.command()
@click.pass_context
@click.option("--retention", type=int, required=True, help="How long to hold AMIs for")
@click.option("--exceptional-amis", help="List of AMI names to always keep just the latest version of (useful for base images)")
@click.option("--launch-templates", help="List of Launch Templates to check AMI usage against. If AMI appears in latest version, it will be spared")
def ami(ctx, retention, exceptional_amis, launch_templates):
    from .ami import cleanup_amis
    region = ctx.obj.get('region')
    not_dry_run = ctx.obj.get('not_dry_run')
    role_arns = ctx.obj.get('role_arns')
    role_arns = role_arns.split(" ")

    if exceptional_amis:
        exceptional_amis = exceptional_amis.split(" ")
    else:
        exceptional_amis = []

    if launch_templates:
        launch_templates = launch_templates.split(" ")
    else:
        launch_templates = []

    try:
        amis = cleanup_amis.CleanupAMIs(region, role_arns, retention, not_dry_run, exceptional_amis, launch_templates)
        amis.cleanup()
        exit(0)
    except Exception as e:
        logging.error(e)
        exit(1)

@cleanup.command()
@click.pass_context
def ebs(ctx):
    from .ebs import cleanup_volumes
    region = ctx.obj.get('region')
    not_dry_run = ctx.obj.get('not_dry_run')
    role_arns = ctx.obj.get('role_arns')
    role_arns = role_arns.split(" ")

    try:
        volumes = cleanup_volumes.CleanupVolumes(region, role_arns, not_dry_run)
        volumes.cleanup()
        exit(0)
    except Exception as e:
        logging.error(e)
        exit(1)

@cleanup.command()
@click.option("--search-tags", required=True, help="Comma separated list of tags attached to snapshots to be deleted")
@click.pass_context
def rds(ctx, search_tags):
    from .rds import cleanup_snapshots
    region = ctx.obj.get('region')
    not_dry_run = ctx.obj.get('not_dry_run')
    role_arns = ctx.obj.get('role_arns')
    role_arns = role_arns

    try:
        snapshots = cleanup_snapshots.CleanupSnapshots(region, role_arns, search_tags, not_dry_run)
        snapshots.cleanup()
        exit(0)
    except Exception as e:
        logging.error(e)
        exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="akinaka",
    version="0.5.3",
    python_requires='>=3.3.0',
    author="Afraz",
    author_email="afraz@olindata.com",
    description="OlinData's AWS CLI Extras",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.olindata.com/olindata/akinaka",
    keywords=[],
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={"console_scripts": ["akinaka=akinaka.main:main"]},
    install_requires=[
        'boto3',
        'datetime',
        'click',
        'pyyaml',
        'kubernetes'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    zip_safe=False
)



import click
from akinaka.client.aws_client import AWS_Client
from akinaka.libs import helpers, scan_resources_storage
from akinaka.libs import helpers, kms_share
from time import gmtime, strftime
import logging

aws_client = AWS_Client()
helpers.set_logger()

@click.group()
@click.option("--region", required=True, help="Region your resources are located in")
@click.option("--source-role-arn", required=True, help="ARN of a role the account to back up _from_")
@click.option("--destination-role-arn", required=True, help="ARN of an assumable role in the account to back up _to_")
@click.option("--dry-run", is_flag=True, help="Don't back anything up, just list would be backed up")
@click.pass_context
def dr(ctx, region, source_role_arn, destination_role_arn, dry_run):
    """
    Disaster recovery subcommand. Does nothing by itself except pass the global options through to it's
    subcommands via ctx
    """

    ctx.obj = {
        'region': region,
        'source_role_arn': source_role_arn,
        'destination_role_arn': destination_role_arn,
        'dry_run': dry_run,
        'log_level': ctx.obj.get('log_level')
    }

    pass

def get_shared_kms_key(region, source_role_arn, source_account, destination_account):
    """
    Create and return shared KMS account between [source_account] and [destination_account]
    """

    kms_sharer = kms_share.KMSShare(
        region = region,
        assumable_role_arn = source_role_arn,
        share_from_account = source_account,
        share_to_account = destination_account
    )

    return kms_sharer.get_kms_key(source_account)

def create_kms_key(region, assumable_role_arn):
    """
    Search for a key name that should exists if this has been run before. If not found,
    create it. In both cases, return the key.
    """

    kms_client = aws_client.create_client('kms', region, assumable_role_arn)
    key_alias = "alias/Akinaka"

    try:
        kms_key = kms_client.describe_key(KeyId=key_alias)
        logging.info("Found key: {}".format(kms_key['KeyMetadata']['Arn']))
    except kms_client.exceptions.NotFoundException:
        kms_key = kms_client.create_key()
        logging.info("No existing key found, so we created one: {}".format(kms_key['KeyMetadata']['Arn']))

        kms_client.create_alias(
            AliasName=key_alias,
            TargetKeyId=kms_key['KeyMetadata']['Arn']
        )

    return kms_key


@dr.command()
@click.pass_context
@click.option("--take-snapshot", is_flag=True, help="Boolean, default false. Take a live snapshot now, or take the existing latest snapshot. Relevant only for RDS")
@click.option("--names", required=False, help="Comma separated list in quotes of DB/S3 names to transfer")
@click.option("--service", type=click.Choice(['rds', 'aurora', 's3']), required=False, help="The service to transfer backups for. Defaults to all (RDS, S3)")
@click.option("--retention", required=False, help="Number of days of backups to keep")
@click.option("--rotate", is_flag=True, required=False, help="Only rotate backups so [retention] number of days is kept, don't do any actual backups. Relevant for RDS only")
@click.option("--keep", required=False, help="Comma separated list in quotes. Do not delete these snapshot IDs as part of the rotation policy.")
def transfer(ctx, take_snapshot, names, service, retention, keep, rotate):
    """
    Creates and passes shared KMS keys to the subcommands which wish to tranfer data between eachother.

    Backup [service] from owning account of [ctx.source_role_arn] to owning account
    of [ctx.destination_role_arn].
    """

    region = ctx.obj.get('region')
    source_role_arn = ctx.obj.get('source_role_arn')
    destination_role_arn = ctx.obj.get('destination_role_arn')
    dry_run = ctx.obj.get('dry_run')

    source_sts_client = aws_client.create_client('sts', region, source_role_arn)
    source_account = source_sts_client.get_caller_identity()['Account']
    destination_sts_client = aws_client.create_client('sts', region, destination_role_arn)
    destination_account = destination_sts_client.get_caller_identity()['Account']

    source_kms_key = get_shared_kms_key(region, source_role_arn, source_account, destination_account)
    destination_kms_key = create_kms_key(region, destination_role_arn)

    if service == 'rds':
        if names:
            db_names = [names.replace(' ','')]
        else:
            scanner = scan_resources_storage.ScanResources(region, source_role_arn)
            db_names = scanner.scan_rds_instances()['db_names']

        if keep:
            keep = [keep.replace(' ','')]

        rds(
            dry_run,
            region,
            source_role_arn,
            destination_role_arn,
            take_snapshot,
            db_names,
            source_kms_key,
            destination_kms_key,
            source_account,
            destination_account,
            retention,
            keep,
            rotate)

    if service == 'aurora':
        if names:
            db_names = [names.replace(' ','')]
        else:
            scanner = scan_resources_storage.ScanResources(region, source_role_arn)
            db_names = scanner.scan_rds_aurora()['aurora_names']

        rds(
            dry_run,
            region,
            source_role_arn,
            destination_role_arn,
            take_snapshot,
            db_names,
            source_kms_key,
            destination_kms_key,
            source_account,
            destination_account,
            retention,
            keep,
            rotate)

    if service == 's3':
        if names:
            names = [names.replace(' ','')]
        else:
            scanner = scan_resources_storage.ScanResources(region, source_role_arn)
            names = scanner.scan_s3()['s3_names']

        s3(
            dry_run,
            region,
            source_role_arn,
            destination_role_arn,
            names,
            source_kms_key,
            destination_kms_key,
            retention
        )

def s3(
        dry_run,
        region,
        source_role_arn,
        destination_role_arn,
        names,
        source_kms_key,
        destination_kms_key,
        retention):
    """ Call the S3 class to make backups of S3 buckets """

    logging.info("Will attempt to backup the following S3 buckets, unless this is a dry run:")
    logging.info(names)

    if dry_run:
        exit(0)

    retention = retention or 7

    from .s3 import transfer_s3
    s3 = transfer_s3.TransferS3(
        region=region,
        source_role_arn=source_role_arn,
        destination_role_arn=destination_role_arn,
        source_kms_key=source_kms_key,
        destination_kms_key=destination_kms_key,
        retention=retention
    )

    s3.main(names)

def rds(
    dry_run,
    region,
    source_role_arn,
    destination_role_arn,
    take_snapshot,
    db_names,
    source_kms_key,
    destination_kms_key,
    source_account,
    destination_account,
    retention,
    keep,
    rotate):
    """
    Call the RDS class to transfer snapshots
    """

    logging.info("Will attempt to backup the data for following RDS instances, unless this is a dry run:")
    logging.info(db_names)

    if dry_run:
        exit(0)

    from .rds import transfer_snapshot
    rds = transfer_snapshot.TransferSnapshot(
        region=region,
        source_role_arn=source_role_arn,
        destination_role_arn=destination_role_arn,
        source_kms_key=source_kms_key,
        destination_kms_key=destination_kms_key
    )

    retention = retention or 7

    if rotate:
        for db_name in db_names:
            rds.rotate_snapshots(retention, db_name, keep)
        exit()

    rds.transfer_snapshot(
        take_snapshot=take_snapshot,
        db_names=db_names,
        source_account=source_account,
        destination_account=destination_account,
        keep=keep,
        retention=retention
    )



import click

# This is the Click() group that's imported into the CLI at the top level
@click.group()
@click.option("--region", required=True, help="Region your resources are located in. N.B. Currently, only us-east-1 supports estimates")
@click.option("--role-arn", required=True, help="Role ARN which contains necessary assume permissions")
@click.pass_context
def container(ctx, region, role_arn):
    ctx.obj = {'region': region, 'role_arn': role_arn}
    pass


@container.command(name="get-ecr-login")
@click.pass_context
@click.option("--registry", required=True, help="Registry you want to retrieve a docker login auth for")
def bill_estimates(ctx, registry):
    region = ctx.obj.get('region')
    role_arn = ctx.obj.get('role_arn')

    from .ecr import ecr_login
    ecr_login.ECRLogin(region, role_arn).get_login(registry)



#!/usr/bin/env python3

import boto3
from time import gmtime, strftime

class AWS_Client():

    def create_client(self, service, region, role_arn, valid_for=None):
        """
        Takes service, region, role_arn, and optionally valid_for (duration in seconds),
        and returns a boto3 client for that service, using that role_arn (with assume role)
        """

        client_options = {
            'region_name': region
        }

        sts_client = boto3.client('sts', region_name=region)

        credentials = sts_client.assume_role(
            RoleArn=role_arn,
            RoleSessionName="akinaka-{}".format(strftime("%Y%m%d%H%M%S", gmtime())),
            DurationSeconds=valid_for or 900
        )

        client_options['aws_access_key_id'] = credentials['Credentials']['AccessKeyId']
        client_options['aws_secret_access_key'] = credentials['Credentials']['SecretAccessKey']
        client_options['aws_session_token'] = credentials['Credentials']['SessionToken']

        return boto3.client(service, **client_options)

    def create_resource(self, service, region, role_arn, valid_for=None):
        """
        Takes service, region, role_arn, and optionally valid_for (duration in seconds),
        and returns a boto3 service object for that service, using that role_arn (with assume role)
        """

        client_options = {
            'region_name': region
        }

        sts_client = boto3.client('sts', region_name=region)

        credentials = sts_client.assume_role(
            RoleArn=role_arn,
            RoleSessionName="akinaka-{}".format(strftime("%Y%m%d%H%M%S", gmtime())),
            DurationSeconds=valid_for or 900
        )

        client_options['aws_access_key_id'] = credentials['Credentials']['AccessKeyId']
        client_options['aws_secret_access_key'] = credentials['Credentials']['SecretAccessKey']
        client_options['aws_session_token'] = credentials['Credentials']['SessionToken']

        return boto3.resource(service, **client_options)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Scans services that may contain data, and returns a list with information on any storage found:

FIXME: EFS is currently out of scope, due to it not being treated as a native AWS service.
       i.e. There is no way to talk to objects stored in EFS outside of a VPC without peering.

[
    {
        "rds_arns": [
            "rds_arn_a",
            "rds_arn_b"
        ],
        "aurora_arns": [
            "rds_arn_a",
            "rds_arn_b"
        ],
        "s3_arns": [
            "s3_arn_a",
            "s3_arn_b"
        ]
    }
]
"""

import boto3
from akinaka.client.aws_client import AWS_Client

aws_client = AWS_Client()

class ScanResources():
    def __init__(self, region, role_arn):
        self.region = region
        self.role_arn = role_arn

    def scan_all(self):
        """ Scan all resource types in scope, and return separate lists for each """

        rds_instance_names = self.scan_rds_instances()
        rds_aurora_names = self.scan_rds_aurora()
        s3_arns = self.scan_s3()
        all_arns = { **rds_instance_names, **rds_aurora_names, **s3_arns }

        return all_arns

    def scan_rds_instances(self):
        """ Return list of ARNs for all RDS DB instances """

        rds_client = aws_client.create_client('rds', self.region, self.role_arn)
        response = rds_client.describe_db_instances()['DBInstances']
        names = [db['DBInstanceIdentifier'] for db in response if 'DBClusterIdentifier' not in db.keys()]

        return { 'db_names': names }

    def scan_rds_aurora(self):
        """ Return list of ARNs for all RDS Aurora objects """

        rds_client = aws_client.create_client('rds', self.region, self.role_arn)
        response = rds_client.describe_db_clusters()['DBClusters']
        names = [db['DBClusterIdentifier'] for db in response]

        return { 'aurora_names': names }

    def scan_s3(self):
        """ Return list of ARNs for all S3 buckets """

        s3_client = aws_client.create_client('s3', self.region, self.role_arn)
        names = [bucket['Name'] for bucket in s3_client.list_buckets()['Buckets']]

        return { 's3_names': names }



from datetime import timedelta, timezone, datetime
from time import strftime
import pprint
import logging

# Take a duration in seconds and work out the datetime value for the datetime at that date and time ago
def datetime_this_seconds_ago(duration):
    return (datetime.now(timezone.utc) + timedelta(seconds=-duration))

def seconds_from_hours(hours):
    return (60*60)*hours

def log(message):
  print(message)

def set_logger(level='INFO'):
    return logging.basicConfig(level=level, format='%(asctime)s [%(levelname)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S%z')



#!/usr/bin/env python3

from time import sleep
from akinaka.libs import helpers
from akinaka.libs import exceptions
import yaml
import logging

class UpdateDeployment():
    """Modify a Kubernetes YAML deployment spec"""

    def __init__(self, applications, new_image=None, new_tag=None, file_paths=None, dry_run=False):
        self.applications = [x.strip() for x in applications.split(',')]
        self.new_image = new_image
        self.new_tag = new_tag
        self.file_paths = [x.strip() for x in file_paths.split(',')]
        self.dry_run = dry_run

        if not self.new_image and not self.new_tag:
            logging.error("At least --new-image or --new-tag need to be given")
            exit(1)

    def read_specs(self, file_path):
        yammies = []

        try:
            with open(file_path, 'r') as stream:
                try:
                    for yammy in yaml.safe_load_all(stream):
                        yammies.append(yammy)
                except yaml.YAMLError as exception:
                    logging.error(exception)
                    exit(1)

            return yammies
        except Exception as exception:
            logging.error(exception)
            exit(1)

    def update_spec(self, file_path):
        specs = self.read_specs(file_path)

        for spec in specs:
            if spec['kind'] == 'Deployment' or spec['kind'] == 'Pod':
                spec_path = spec['spec']['template']['spec']['containers']
            elif spec['kind'] == 'CronJob':
                spec_path = spec['spec']['jobTemplate']['spec']['template']['spec']['containers']
            else:
                logging.error("Can't handle spec file type of {}".format(spec['kind']))
                exit(1)

            for container in spec_path:
                if container['name'] in self.applications:
                    if self.new_image:
                        image = self.new_image
                    else:
                        image = container['image'].split(":")[0]

                    if self.new_tag:
                        tag = self.new_tag
                    else:
                        tag = container['image'].split(":")[-1]

                    container['image'] = "{image}:{tag}".format(image = image, tag = tag)
                else:
                    logging.warning("Didn't find every container searched for from the array {}, and didn't change any files".format(self.applications))

            # spec['metadata']['annotations']['kubernetes.io/change-cause'] = self.new_tag


        return specs

    def write_new_spec(self):
        for file_path in self.file_paths:

            new_spec = yaml.dump_all(self.update_spec(file_path), default_flow_style=False)

            if not self.dry_run:
                spec_file = open(file_path, "w")
                spec_file.write(new_spec)
                spec_file.close()
            else:
                print(new_spec)



import boto3
from akinaka.client.aws_client import AWS_Client
import akinaka.libs.helpers as helpers
from datetime import datetime, timezone, timedelta

aws_client = AWS_Client()

class CostExplorer():
    def __init__(self, region, role_arn):
        self.costexplorer_client = aws_client.create_client('ce', region, role_arn)

    def get_bill_estimates(self, from_days_ago):
        days_ago = 0 or int(from_days_ago)

        if days_ago > 0:
            end = datetime.now().strftime("%Y-%m-%d")
            datetime_days_ago = datetime.now() - timedelta(days=days_ago + 1)
            start = datetime_days_ago.strftime("%Y-%m-%d")
        else:
            start = datetime.now().strftime("%Y-%m-%d")
            datetime_days_ago = datetime.now() + timedelta(days=1)
            end = datetime_days_ago.strftime("%Y-%m-%d")

        return self.costexplorer_client.get_cost_and_usage(
            TimePeriod={
                'Start': start,
                'End': end,
            },
            Granularity='DAILY',
            Metrics=[
                'UnblendedCost'
            ]
        )
